﻿Unit awsassist;

{**************************************************************************

   SHA256 and HMAC-SHA256 for Amazon Web Service Assist

   MD5 from rfc1231.txt

   Copyright (C) 2002 Go Maeda <g_maeda@mars.dti.ne.jp>
   This work is derived from the RSA Data Security, Inc. MD5 Message-Digest
   Algorithm.

   Convert for Delphi by Ayakawa,Marinosuke

   2000/11/17 1st. convert to delphi
   2004/07/04 2nd. change param string to pbyte
   2006/09/03 3rd. use $Q-/$Q+ for D7

   Copyright (C) 2009 Akira,Otaki
     Create SHA256 and HMAC-SHA256 from MD5 Message-Digest

   2009/06/02 Modify to SHA256 and HMAC-SHA256.
   2009/06/03 Add signrequest function.
   2009/06/05 bugfix sort order of aws parameters.

   このユニットはご自由に使用してください。
   配布・転載もご自由にどうぞ。
   なお、使用する場合はご自身の責任にてどうぞ。作者は一切責任を持ちません。

 **************************************************************************}

{$IFDEF FPC}
  {$mode objfpc}{$H+}
{$ENDIF}

interface

uses  windows,sysutils,classes;

type
  TARYPARAM8 = array[0..7] of DWORD;
  PARYPARAM8 = ^TARYPARAM8;
  THASH256 = array[0..31] of byte;

function CalcSHA256(ASrc:PByte; ASize:integer):string;
function CalcSHA256HM(ASrc:PByte; ASize:integer;KSrc:PByte; KSize:integer):string;
function signrequest(url:String;key:String;key1,key2:PARYPARAM8): String;

implementation

type
  UINT4 = DWORD;

  TARYUINT8 = array[0..7] of UINT4;
  TARYBYTE64 = array[0..63] of byte;

  TSHA256_CTX = packed record
    state : TARYUINT8;
    count : array[0..1] of UINT4;
    buffer : TARYBYTE64;
  end;

  TSHA256HM_CTX = packed record
    SHA256Context : TSHA256_CTX;
    key0 : TARYBYTE64;
    FUseKey:Boolean;
    flag:Boolean;
  end;

const
  PADDING:AnsiString = #$80#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0+
            #0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0+
            #0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0+
            #0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0;

  sha256_kt: array[0..63] of DWORD =
    ($428a2f98,$71374491,$b5c0fbcf,$e9b5dba5,
     $3956c25b,$59f111f1,$923f82a4,$ab1c5ed5,
     $d807aa98,$12835b01,$243185be,$550c7dc3,
     $72be5d74,$80deb1fe,$9bdc06a7,$c19bf174,
     $e49b69c1,$efbe4786,$0fc19dc6,$240ca1cc,
     $2de92c6f,$4a7484aa,$5cb0a9dc,$76f988da,
     $983e5152,$a831c66d,$b00327c8,$bf597fc7,
     $c6e00bf3,$d5a79147,$06ca6351,$14292967,
     $27b70a85,$2e1b2138,$4d2c6dfc,$53380d13,
     $650a7354,$766a0abb,$81c2c92e,$92722c85,
     $a2bfe8a1,$a81a664b,$c24b8b70,$c76c51a3,
     $d192e819,$d6990624,$f40e3585,$106aa070,
     $19a4c116,$1e376c08,$2748774c,$34b0bcb5,
     $391c0cb3,$4ed8aa4a,$5b9cca4f,$682e6ff3,
     $748f82ee,$78a5636f,$84c87814,$8cc70208,
     $90befffa,$a4506ceb,$bef9a3f7,$c67178f2);

  sha256_init: array[0..7] of DWORD =
    ($6a09e667,$bb67ae85,$3c6ef372,$a54ff53a,
     $510e527f,$9b05688c,$1f83d9ab,$5be0cd19);

  Base64 : AnsiString = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'+
                        'abcdefghijklmnopqrstuvwxyz'+
                        '0123456789+/';

{$Q-}

function base64_encode(ASrc:PByte;ASize:integer):String;
var
  wk : DWORD;
  a,b,c,d:Integer;
  i:Integer;
  Ans:String;
  pb:PByte;
begin
  i:=0;
  Ans:='';
  pb:=ASrc;
  while i<ASize do
  begin
    wk:=(pb^ shl 16);inc(pb);
    if i+1<ASize then
    begin
      wk:=wk or (pb^ shl 8);inc(pb);
    end;
    if i+2<ASize then
    begin
      wk:=wk or pb^;inc(pb);
    end;

    a:=(wk shr 18) and $3f;
    b:=(wk shr 12) and $3f;
    c:=(wk shr 6) and $3f;
    d:=wk and $3f;
    if i+2<ASize then
    begin
      Ans:=Ans+Base64[a+1]+Base64[b+1]+Base64[c+1]+Base64[d+1];
    end else begin
      Ans:=Ans+Base64[a+1]+Base64[b+1];
      if i+1 < ASize then
        Ans:=Ans+Base64[c+1]
      else
        Ans:=Ans+'=';
      if i+2 < ASize then
        Ans:=Ans+Base64[d+1]
      else
        Ans:=Ans+'=';
    end;
    inc(i,3);
  end;
  result:=Ans;
end;

function urlencode(Data:AnsiString):String;
var
  Ans:String;
  i:Integer;
  c:Char;
begin
  Ans:='';

  for i:=1 to Length(Data) do
  begin
    c:=Data[i];
    if (c>='0') And (c<='9') then
      Ans:=Ans+c
    else
    if (c>='a') And (c<='z') then
      Ans:=Ans+c
    else
    if (c>='A') And (c<='Z') then
      Ans:=Ans+c
    else
    if (c='-') or (c='.') or (c='_') then
      Ans:=Ans+c
    else begin
      Ans:=Ans+'%'+UpperCase(IntToHex(Ord(c) And $ff,2));
    end;
  end;
  result:=Ans;
end;

function urldecode(Data:String):String;
var
  Ans:String;
  i:Integer;
  c:String;
begin
  Ans:='';
  i:=1;

  while i <=  Length(Data) do
  begin
    c:=Data[i];
    if (c='%') And (i+2<=Length(Data)) then
    begin
      c:=Data[i+1];
      c:=c+Data[i+2];
      Ans:=Ans+Char(StrToInt('$'+c) And $ff);
      inc(i,2);
    end else
      Ans:=Ans+c;
    inc(i);
  end;
  result:=Ans;
end;

function sha_rotr(x,n:DWORD):DWORD;
begin
  result:= (x SHR n) OR (x SHL (32-n));
end;

function sha_shr(x,n:DWORD):DWORD;
begin
  result:= (x SHR n);
end;

function sha_SL256_0(x:DWORD):DWORD;
begin
  result:= sha_rotr(x,2) XOR sha_rotr(x,13) XOR sha_rotr(x,22);
end;

function sha_SL256_1(x:DWORD):DWORD;
begin
  result:=sha_rotr(x,6) XOR sha_rotr(x,11) XOR sha_rotr(x,25);
end;

function sha_SS256_0(x:DWORD):DWORD;
begin
  result:=sha_rotr(x,7) XOR sha_rotr(x,18) XOR sha_shr(x,3);
end;

function sha_SS256_1(x:DWORD):DWORD;
begin
  result:=sha_rotr(x,17) XOR sha_rotr(x,19) XOR sha_shr(x,10);
end;

function sha_Ch32(x,y,z:DWORD):DWORD;
begin
  result:= (x AND y) XOR ((NOT x) AND z);
end;

function sha_Maj32(x,y,z:DWORD):DWORD;
begin
  result:= (x AND y) XOR (x AND z) XOR (y AND z);
end;

// for (i=0; i<cpylen; i++) context->buffer[bufidx+i]=src[srcidx+i]
procedure SHA256_memcpy(var context:TSHA256_CTX; bufidx:integer; src:PByte; srcidx:integer; cpylen:integer);
var
  i : integer;
  p : PByte;
begin
  // srcidx 0-origin. but string is 1-origin => 1+srcidx+i
  //FOR i:=0 TO cpylen-1 DO context.buffer[bufidx+i]:=BYTE(ord(src[1+srcidx+i]));
  p:=src;
  Inc(p,srcidx);
  FOR i:=0 TO cpylen-1 DO BEGIN
    context.buffer[bufidx+i]:=p^; Inc(p);
  END;
end;

// SHA256 basic transformation. Transforms state based on block.
procedure SHA256Transform(var state:TARYUINT8; block:TARYBYTE64);
var
  a,b,c,d,e,f,g,h : UINT4;
  x : array[0..15] of UINT4;
  j : Integer;
  t1,t2 : UINT4;
  {xs : array[0..15] of String;
  hs : array[0..7] of String;}

  procedure Decode;
  var
    i,j : integer;
  begin
    i:=0; j:=0;
    while j<64 do
    begin
      x[i]:=block[j+3] OR (block[j+2] SHL 8) OR (block[j+1] SHL 16) OR (block[j+0] SHL 24);
      INC(i); INC(j,4);
    end;
  end;
begin
  a:=state[0]; b:=state[1]; c:=state[2]; d:=state[3];
  e:=state[4]; f:=state[5]; g:=state[6]; h:=state[7];

  Decode;

  {for j:=0 to 15 do
  begin
    xs[j]:=IntToHex(x[j],8);
  end;}

  for j := 0 to 63 do
  begin
    if j >= 16 then
    begin
      x[j MOD 16] := sha_SS256_1(x[(j-2) MOD 16]) + x[(j-7) MOD 16] + sha_SS256_0(x[(j-15) MOD 16]) + x[(j-16) MOD 16];
    end;

    t1 := h + sha_SL256_1(e) + sha_Ch32(e,f,g) + x[j MOD 16] + sha256_kt[j];
    t2 := sha_SL256_0(a) + sha_Maj32(a,b,c);
    h := g;
    g := f;
    f := e;
    e := d+t1;
    d := c;
    c := b;
    b := a;
    a := t1+t2;

    {hs[0] := IntToHex(a,8);
    hs[1] := IntToHex(b,8);
    hs[2] := IntToHex(c,8);
    hs[3] := IntToHex(d,8);
    hs[4] := IntToHex(e,8);
    hs[5] := IntToHex(f,8);
    hs[6] := IntToHex(g,8);
    hs[7] := IntToHex(h,8);}

  end;

  inc(state[0],a);
  inc(state[1],b);
  inc(state[2],c);
  inc(state[3],d);
  inc(state[4],e);
  inc(state[5],f);
  inc(state[6],g);
  inc(state[7],h);

  // Zeroize sensitive information.
  FillChar(x[0],sizeof(x),0); // SHA256_memset
end;

procedure SHA256Init(var sha256ctx:TSHA256_CTX);
var
  i : integer;
begin
  WITH sha256ctx DO BEGIN
    count[0]:=0;
    count[1]:=0;
    // Load magic initialization constants.
    for i:=0 to 7 do
    begin
      state[i]:=sha256_init[i];
    end;
  END;
end;

procedure SHA256Update(var context:TSHA256_CTX; inputdata:PByte; inputlen:DWORD);
var
  i,index,partLen : DWORD;
  work : TARYBYTE64;
  pB : PByte;
begin
  // Compute number of bytes mod 64
  index := (context.count[0] SHR 3) AND $3F;

  // Update number of bits
  INC(context.count[0],(inputlen SHL 3));
  IF context.count[0]<(inputlen SHL 3) THEN INC(context.count[1]);
  INC(context.count[1],(inputlen SHR 29));

  partLen:=64-index;

  // Transform as many times as possible
  IF inputlen >= partLen THEN BEGIN
    SHA256_memcpy(context,index,inputdata,0,partLen);
    SHA256Transform(context.state,context.buffer);
    i:=partlen;
    WHILE (i+63)<inputlen DO BEGIN
      pB:=inputdata; Inc(pB,i);
      Move(pB^,work[0],64);
      SHA256Transform(context.state,work);
      INC(i,64);
    END;
    index:=0;
  END ELSE
    i:=0;

  // Buffer remaining input
  SHA256_memcpy(context,index,inputdata,i,inputlen-i);
end;

procedure  SHA256FinalRaw(var context:TSHA256_CTX;var hash:THASH256);
var
  bits : array[0..7] of BYTE;
  work : AnsiString;
  index,padlen : DWORD;
  i,j : integer;
  procedure Encode8;
  begin
    i:=1; j:=0;
    WHILE j<8 DO BEGIN
      bits[j+3]:=context.count[i] AND $FF;
      bits[j+2]:=(context.count[i] SHR 8) AND $FF;
      bits[j+1]:=(context.count[i] SHR 16) AND $FF;
      bits[j+0]:=(context.count[i] SHR 24) AND $FF;
      dec(i); inc(j,4);
    END;
  end;
  procedure Encode32;
  begin
    i:=0; j:=0;
    WHILE j<32 DO BEGIN
      hash[j+0]:=(context.state[i] SHR 24) AND $FF;
      hash[j+1]:=(context.state[i] SHR 16) AND $FF;
      hash[j+2]:=(context.state[i] SHR 8) AND $FF;
      hash[j+3]:=context.state[i] AND $FF;
      INC(i); INC(j,4);
    END;
  end;
begin
  // Save number of bits
  Encode8;

  // Pad out to 56 mod 64
  index:=(context.count[0] SHR 3) AND $3F;
  IF index<56 THEN padlen:=56-index ELSE padlen:=120-index;
  SHA256Update(context,PByte(@PADDING[1]),padlen);

  // Append length (before padding)
  work:='        '; // need 8byte
  Move(bits[0],work[1],8);
  SHA256Update(context,pByte(@work[1]),8);

  // Store state in digest
  Encode32;

  // Zeroize sensitive information.
  FillChar(context,sizeof(context),0);
end;

function  SHA256Final(var context:TSHA256_CTX):string;
var
  hash : THASH256;
  j : integer;
  function Encode32:string;
  begin
    result:='';
    j:=0;
    WHILE j<32 DO BEGIN
      result:=result+IntToHex(hash[j] AND $FF,2);
      inc(j);
    END;
    result:=LowerCase(result);
  end;
begin
  SHA256FinalRaw(context,hash);

  // Store state in digest of hexadecimal
  result:=Encode32;
end;

procedure SHA256HMInit(var sha256hmctx:TSHA256HM_CTX;KSrc:PByte;KSize:Integer);
const
  blocksize:integer = 64;
var
  context : TSHA256_CTX;
  hash : THASH256;
begin
  with sha256hmctx do begin
    // Load magic initialization constants.
    if KSize > blocksize then
    begin
      SHA256Init(context);
      SHA256Update(context,KSrc,KSize);
      SHA256FinalRaw(context,hash);
      KSize:=blocksize-32;
      Move(hash[0],key0[0],KSize);
    end else begin
      Move(KSrc^,key0[0],KSize);
    end;
    // padding 0x00 to blocksize
    FillChar(key0[KSize],blocksize-KSize,0);
    SHA256Init(SHA256Context);
    flag:=false;
  end;
end;

procedure SHA256HMSetCryptKey(var sha256hmctx:TSHA256HM_CTX;nblks:Integer;var param8:TARYPARAM8);
var
  i:integer;
  step2 : THASH256;
  flg : boolean;
begin
  flg:=false;
  if not ((sha256hmctx.SHA256Context.count[0]=0) and (sha256hmctx.SHA256Context.count[1]=0)) then
  begin
    SHA256FinalRaw(sha256hmctx.SHA256Context,step2);
    SHA256Init(sha256hmctx.SHA256Context);
    flg:=true;
  end;
  for i:=0 to 7 do
  begin
    sha256hmctx.SHA256Context.state[i]:=param8[i];
  end;
  sha256hmctx.SHA256Context.Count[0]:=nblks*512;
  sha256hmctx.SHA256Context.Count[1]:=0;
  if flg then
  begin
    SHA256Update(sha256hmctx.SHA256Context,@step2[0],sizeof(step2));
    sha256hmctx.flag:=true;
  end;
end;

procedure SHA256HMGetCryptKey(var sha256hmctx:TSHA256HM_CTX;var param8:TARYPARAM8);
  procedure step1;
  const
    blocksize:integer = 64;
  var
    ipadkey : Array[0..63] of BYTE;
    i:Integer;
  begin
    for i:=0 to blocksize-1 do
    begin
      ipadkey[i]:=sha256hmctx.key0[i] XOR $36;
    end;
    SHA256Update(sha256hmctx.SHA256Context,@ipadkey[0],blocksize);
  end;
  procedure step3(var step2:THASH256);
  const
    blocksize:integer = 64;
  var
    opadkey : Array[0..63] of BYTE;
    i:Integer;
  begin
    for i:=0 to blocksize-1 do
    begin
      opadkey[i]:=sha256hmctx.key0[i] XOR $5c;
    end;

    SHA256FinalRaw(sha256hmctx.SHA256Context,step2);
    SHA256Init(sha256hmctx.SHA256Context);
    SHA256Update(sha256hmctx.SHA256Context,@opadkey[0],blocksize);
  end;
var
  i:integer;
  step2 : THASH256;
  flg : boolean;
begin
  flg:=false;
  if (sha256hmctx.SHA256Context.count[0]=0) and (sha256hmctx.SHA256Context.count[1]=0) then
  begin
    step1;
  end else
  if not ((sha256hmctx.SHA256Context.count[0]=0) and (sha256hmctx.SHA256Context.count[1]=0)) then
  begin
    step3(step2);
    flg:=true;
  end;
  for i:=0 to 7 do
  begin
    param8[i]:=sha256hmctx.SHA256Context.state[i];
  end;
  if flg then
  begin
    SHA256Update(sha256hmctx.SHA256Context,@step2[0],sizeof(step2));
    sha256hmctx.flag:=true;
  end;
end;

procedure SHA256HMUpdate(var sha256hmctx:TSHA256HM_CTX;ASrc:PByte;ASize:Integer);
  procedure step1;
  const
    blocksize:integer = 64;
  var
    ipadkey : Array[0..63] of BYTE;
    i:Integer;
  begin
    for i:=0 to blocksize-1 do
    begin
      ipadkey[i]:=sha256hmctx.key0[i] XOR $36;
    end;
    SHA256Update(sha256hmctx.SHA256Context,@ipadkey[0],blocksize);
  end;
begin
  if (sha256hmctx.SHA256Context.count[0]=0) and (sha256hmctx.SHA256Context.count[1]=0) then
  begin
    step1;
  end;
  SHA256Update(sha256hmctx.SHA256Context,ASrc,ASize);
end;

procedure SHA256HMFinalRaw(var sha256hmctx:TSHA256HM_CTX;var hash:THASH256);
  procedure step3(var step2:THASH256);
  const
    blocksize:integer = 64;
  var
    opadkey : Array[0..63] of BYTE;
    i:Integer;
  begin
    for i:=0 to blocksize-1 do
    begin
      opadkey[i]:=sha256hmctx.key0[i] XOR $5c;
    end;

    SHA256FinalRaw(sha256hmctx.SHA256Context,step2);
    SHA256Init(sha256hmctx.SHA256Context);
    SHA256Update(sha256hmctx.SHA256Context,@opadkey[0],blocksize);
  end;
var
  step2 : THASH256;
begin
  if not sha256hmctx.flag then
  begin
    step3(step2);
    SHA256Update(sha256hmctx.SHA256Context,@step2[0],sizeof(step2));
  end;
  SHA256FinalRaw(sha256hmctx.SHA256Context,hash);
end;

function  SHA256HMFinal(var sha256hmctx:TSHA256HM_CTX):string;
var
  hash : THASH256;
  j : integer;
  function Encode32:string;
  begin
    result:='';
    j:=0;
    WHILE j<32 DO BEGIN
      result:=result+IntToHex(hash[j] AND $FF,2);
      inc(j);
    END;
    result:=LowerCase(result);
  end;
begin
  SHA256HMFinalRaw(sha256hmctx,hash);

  // Store state in digest of hexadecimal
  //result:=Encode32;
  result:=base64_encode(@hash[0],sizeof(hash));
end;

function CalcSHA256HM(ASrc:PByte; ASize:integer;KSrc:PByte; KSize:integer):string;
var
  context : TSHA256HM_CTX;
begin
  SHA256HMInit(context,KSrc,KSize);
  SHA256HMUpdate(context,ASrc,ASize);
  Result:=SHA256HMFinal(context);
end;

function CalcSHA256(ASrc:PByte; ASize:integer):string;
var
  context : TSHA256_CTX;
begin
  SHA256Init(context);
  SHA256Update(context,ASrc,ASize);
  Result:=SHA256Final(context);
end;

function CompareByteOrder(List:TStringList;Index1:Integer;Index2:Integer):Integer;
var
  s1,s2:String;
begin
  s1:=List.strings[Index1];
  s2:=List.strings[Index2];
  if s1=s2 then
    Result:=0
  else
  if s1 < s2 then
    Result:=-1
  else
    Result:=1;
end;

function signrequest(url:String;key:String;key1,key2:PARYPARAM8): String;
const
  skipblock:integer = 1;
  dummy:String = 'dummy';
var
  protocol:String;
  host:String;
  port:String;
  method:String;
  path:String;
  remain:String;
  idx,idx1,idx2:Integer;
  ParamList:TStringList;
  QueryList:TStringList;
  i:Integer;
  bTimestampExist:boolean;
  sName,sValue:String;
  s:String;
  signstring,paramstring,resulturl,signaturestring:String;
  context:TSHA256HM_CTX;
  systime:TSystemTime;
  datetime:TDatetime;
begin
  result:='';
  host:='';
  protocol:='http';
  port:='80';
  path:='/';
  remain:=url;
  method:='GET';
  bTimestampExist:=false;


  ParamList:=TStringList.Create;
  try
  ParamList.Clear;
  ParamList.CaseSensitive:=true;
  // parse url

  // analyze protocol
  if Copy(remain,1,7)='http://' then
  begin
    protocol:='http';
    port:='80';
    remain:=Copy(remain,8,Length(remain));
  end else
  if Copy(remain,1,8)='https://' then
  begin
    protocol:='https';
    port:='433';
    remain:=Copy(remain,9,Length(remain));
  end else begin
    exit;
  end;

  // separate host/path/parameter
  idx1:=Pos('/',remain);
  idx2:=Pos('?',remain);
  if ((idx1>0) And (idx2 > 0) And (idx2<idx1)) then
    idx1:=0;

  // analyze host
  if ((idx1>0) or (idx2>0)) then
  begin
    idx:=idx1;
    if idx = 0 then
      idx:=idx2;
    host:=Copy(remain,1,idx-1);
    remain:=Copy(remain,idx,length(remain));
  end else begin
    host:=remain;
    remain:='';
  end;

  // analyze path
  if idx1 > 0 then
  begin
    if idx2 > 0 then
    begin
      idx:=Pos('?',remain);
      path:=Copy(remain,1,idx-1);
      remain:=Copy(remain,idx+1,Length(remain));
    end else begin
      path:=remain;
      remain:='';
    end;
  end else begin
    path:='/';
  end;

  // analyze query
  if idx2 > 0 then
  begin
    QueryList:=TStringList.Create;
    try
      QueryList.Delimiter:='&';
      QueryList.DelimitedText:=remain;
      for i:=0 to QueryList.Count -1 do
      begin
        s:=QueryList.strings[i];
        idx:=Pos('=',s);
        if idx>0 then
        begin
          sName:=Copy(s,1,idx-1);
          sValue:=Copy(s,idx+1,Length(s));
          // decode parameter
          sName:=urldecode(sName);
          sValue:=urldecode(sValue);
          ParamList.Append(sName+'='+sValue);
          if sName='Timestamp' then
            bTimestampExist:=true;
        end;
      end;
    finally
      QueryList.Free;
    end;
  end;

  // append Timestamp if not exists
  if not bTimestampExist then
  begin
    sName:='Timestamp';
    GetSystemTime(systime);
    datetime:=SystemTimeToDatetime(systime);
    sValue:=FormatDateTime('YYYY-MM-DD"T"HH:NN:SS"Z"',datetime);
    ParamList.Append(sName+'='+sValue);
  end;

  // sort parameter
  ParamList.CustomSort(@CompareByteOrder);

  // remake string and url
  signstring:='';
  signstring:=signstring + method + #10;
  signstring:=signstring + host + #10;
  signstring:=signstring + path + #10;
  paramstring:='';
  for i:=0 to ParamList.Count -1 do
  begin
    s:=ParamList.strings[i];
    idx:=Pos('=',s);
    if idx>0 then
    begin
      sName:=Copy(s,1,idx-1);
      sValue:=Copy(s,idx+1,Length(s));
      sName:=urlencode(sName);
      sValue:=urlencode(sValue);
      if Length(paramstring) > 0 then
      begin
        paramstring:=paramstring + '&';
      end;
      paramstring:=paramstring + sName + '=' + sValue;
    end;
  end;
  signstring:=signstring + paramstring;
  resulturl:='';
  resulturl:=resulturl + 'http://';
  resulturl:=resulturl + host;
  resulturl:=resulturl + path;
  resulturl:=resulturl + '?';
  resulturl:=resulturl + paramstring;

  // call hm-sha256
  SHA256HMInit(context,PByte(key),Length(key));
  if Length(key) = 0 then
  begin
    SHA256HMSetCryptKey(context,skipblock,key1^)
  end else begin
    if key1 <> nil then
      SHA256HMGetCryptKey(context,key1^)
  end;

  SHA256HMUpdate(context,PByte(signstring),Length(signstring));
  if Length(key) = 0 then
  begin
    SHA256HMSetCryptKey(context,1,key2^);
  end else begin
    if key2 <> nil then
      SHA256HMGetCryptKey(context,key2^)
  end;
  signaturestring:=SHA256HMFinal(context);

  // append Signature
  resulturl:=resulturl + '&';
  resulturl:=resulturl + 'Signature=' + urlencode(signaturestring);
  Result:=resulturl;
  finally
    ParamList.Free;
  end;
end;

{$Q+}

end.
